/*
 * LTS.java
 *
 * Created on 23 July 2007, 23:13
 */

package assignment;

//import com.sun.org.apache.bcel.internal.classfile.JavaClass;
import java.util.*;
import java.lang.*;
import java.io.*;

/**
 *
 * @author David
 */
public class LTS {
   
    protected String fileName;
    protected LTSNode startNode;
    protected HashSet<LTSTransition> transitionList;
    
    /** This is used as a parameter to the parallel reach recursive function
     */ 
    private static class parallel_reach_params {
        public LTS ltsA;
        public LTS ltsB;
        public HashSet<String> syncSet;
        
        public HashMap<LTSNode, HashSet<LTSTransition>> nodeMapA;
        public HashMap<LTSNode, HashSet<LTSTransition>> nodeMapB;
        public Hashtable<String,LTSTransition> transitionMapA;
        public Hashtable<String,LTSTransition> transitionMapB;
        
        public Collection<LTSNode> nodesA;
        public Collection<LTSNode> nodesB;
        
        public Collection<LTSTransition> transitionsA;
        public Collection<LTSTransition> transitionsB;
        
        public LTS newLTS;
        
        public HashSet<LTSNode> visitedNodes;
    }
    
    /** Creates a new instance of LTS */
    public LTS() {
        transitionList = new HashSet<LTSTransition>();//new ArrayList<LTSTransition>();
    }
    
    /** Sets the filename in use. This should only really be used by the load function - if you
     * want to save as a different file name use the save(string) function.
     */
    public void setFileName(String filename) {
        fileName = filename;
    }
    
    /** Returns the start node for this LTS
     */
    public LTSNode getStartNode() {
        return startNode;
    }
    
    /** Sets the start node for this LTS
     */
    public void setStartNode(LTSNode start) {
        startNode = start;
    }
    
    /** Returns true if the specified transition exists in this LTS.
     */
    public boolean transitionExists(LTSTransition ltst) {
        return transitionList.contains(ltst);
    }
    
    /** Adds the specified transition to this LTS
     */
    public void addTransition(LTSTransition ltst) {
        if (!transitionList.contains(ltst))
            transitionList.add(ltst);
    }
    
    /** Removes the specified transition from this LTS.
     */
    public void removeTransition(LTSTransition ltst) {
        transitionList.remove(ltst);
    }
      
    /** Removes all transitions from the LTS.
     */
    public void killTransitions() {
        transitionList.clear();
    }
    
    /** Returns a list of all transitions in this LTS.
     */
    public Collection<LTSTransition> getTransitions() {
        return (Collection<LTSTransition>)transitionList.clone();
    }
    
    /** Returns a list of nodes in the LTS.
     * @param includeStart If set to true this function will return the start node
     * even if it is not present in any transitions. If false it will only return
     * the start node if it is present in one or more transitions
     * @returns A list of nodes in the LTS
     */
    public Collection<LTSNode> getNodes(boolean includeStart) {
        Collection<LTSTransition> tl = getTransitions();
        ArrayList<LTSNode> output = new ArrayList<LTSNode>();
        
        HashSet<LTSNode> hs = new HashSet<LTSNode>();
        
        Iterator it = tl.iterator();
        
        while (it.hasNext()) {
            LTSTransition lt = (LTSTransition) it.next();
            
            if (lt.getNode1() != null)
                hs.add(lt.getNode1());
            if (lt.getNode2() != null)
                hs.add(lt.getNode2());
        }
        
        ArrayList<LTSNode> nl = new ArrayList<LTSNode>();
        
        nl.addAll(hs);
        
        if ((getStartNode() != null) && includeStart)
            nl.add(getStartNode());
        
        return nl;
    }
    
    /** This function is for compatibility with older code. It should not be used.
     * It just calls getNodes with includeStart = true.
     * @returns A list of all nodes in the LTS (including the start node even if
     * it isnt a member of a transition.
     */
    public Collection<LTSNode> getNodes() {
        return getNodes(true);
    }
    
    /** Loads an LTS from a text file.
     */
    public static LTS load(String filename) throws IOException {
        System.out.println("LTS: Attempting to load file \"" + filename +"\"");
        
        LTS theLTS = new LTS();
        theLTS.setFileName(filename);
        
        FileInputStream fin;		

        fin = new FileInputStream (filename);
        BufferedReader breader = new BufferedReader(new InputStreamReader(fin));            
        
        String startNodeName = "";
        
        String curlin = breader.readLine();
        
        while (curlin != null) {
            if (curlin.startsWith("#")) {
                // Its a comment, do nothing
            }
            else {
                if (startNodeName == "") {
                    startNodeName = curlin;
                    theLTS.setStartNode(new LTSNode(startNodeName));
                }
                else {
                    String node1;
                    String node2;
                    String name;

                    String[] stringList = curlin.split("/");

                    node1 = stringList[0];
                    if (stringList.length > 1) {
                        name = stringList[1];
                        node2 = stringList[2];

                        theLTS.addTransition(new LTSTransition(new LTSNode(node1),new LTSNode(node2),name));
                    }
                    else {
                        theLTS.addTransition(new LTSTransition(new LTSNode(node1),null,""));
                    }
                }
            }
            
            curlin = breader.readLine();
        }



        fin.close();		

        
        return theLTS;
    }
    
    /** Saves the LTS to the current file ("Save")
     */
    public void save() throws IOException {
        save(getFileName());
    }
    
    /** Saves the LTS to the specified file ("Save As")
     */
    public void save(String filename) throws IOException {
        setFileName(filename);  // Just in case the file name is different.
        FileOutputStream fout;		


        fout = new FileOutputStream (getFileName());
        PrintStream ps = new PrintStream(fout);
        
        // Print the start node:
        ps.println (getStartNode().getName());
        
        // Print the transitions:
        for (LTSTransition t:transitionList) {
            ps.println(t.toString());
        }

        fout.close();		

    }
    
    /** Compares the LTS with another.
     *@returns true if the two LTS are equal.
     */
    public boolean equals(LTS theLTS){
        // Ensure they have the same start nodes
        String lSN = theLTS.getStartNode().getName();
        String lSNN = "";
        if (getStartNode() != null)
            lSNN = getStartNode().getName();
        if (!lSN.equals(lSNN))
            return false;
        
        // If one has more transitions than the other they cant be equal.
        if (theLTS.getTransitions().size() != getTransitions().size())
            return false;
        
        // Ensure they have the same transitions
        for (LTSTransition t:transitionList){
            boolean te = theLTS.transitionExists(t);
            if (!te)
                return false;
        }
        return true;
    }
    
    /** Composes two LTS and returns the result. Clones lts_a and the composes it with lts_b returning the result.
     * @returns The composition of lts_a and lts_b using the given Sync Set.
     */
    public static LTS compose(LTS lts_a, LTS lts_b, HashSet<String> syncSet) {
        
        // We cant have a null SyncSet now can we...
        if (syncSet == null) syncSet = new HashSet<String>();
       
        LTS theNewLTS = new LTS();
        
        // First up, get the nodes from each LTS
        Collection<LTSNode> nA = lts_a.getNodes(false);
        Collection<LTSNode> nB = lts_b.getNodes(false);
        
        Collection<LTSTransition> ts = lts_a.getTransitions();
        
        // In the interest of performance, we must compute a hashtable of transitions.
        Hashtable<String,LTSTransition> htA = new Hashtable<String,LTSTransition>();
        Hashtable<String,LTSTransition> htB = new Hashtable<String,LTSTransition>();
        for (LTSTransition t:ts) {
            htA.put(t.getName(),t);
        }
        for (LTSTransition t:lts_b.getTransitions()) {
            htB.put(t.getName(),t);
        }
        

        
        
        // Now compute the transitions for the transitions from the first LTS and the nodes from the second LTS
        for (LTSTransition t: ts) {
            // Ok, we must check to see if this transition is in both LTS and the sync set.
            if (syncSet.contains(t.getName()) && htB.containsKey(t.getName())) {
                // It is in all 3 sets.
                LTSTransition b = htB.get(t.getName());
                
                LTSTransition lt = new LTSTransition();
                lt.setName(t.getName());
                lt.setNode1(new LTSNode("(" + t.getNode1().getName() + "," + b.getNode1().getName() + ")"));
                lt.setNode2(new LTSNode("(" + t.getNode2().getName() + "," + b.getNode2().getName() + ")"));
                theNewLTS.addTransition(lt);
                continue;
            }
            else {
                // treat it normally.
                if (t.getNode2() != null)
                    for (LTSNode n : nB) {
                        LTSTransition lt = new LTSTransition();
                        lt.setName(t.getName());
                        lt.setNode1(new LTSNode("(" + t.getNode1().getName() + "," + n.getName() + ")"));
                        lt.setNode2(new LTSNode("(" + t.getNode2().getName() + "," + n.getName() + ")"));
                        theNewLTS.addTransition(lt);
                    }
            }
        }
        
        ts = lts_b.getTransitions();
       
        // Now compute the transitions for the transitions from the second LTS and the nodes from the first LTS
        for (LTSTransition t: ts) {
            if (syncSet.contains(t.getName()) && htA.containsKey(t.getName())) {
                // The node is synchronized and has already been added. Nothing to do here...
                continue;
            }
            else {
                // some transitions have null as their second node.
                // These really just represent a single node - not a transition.
                if (t.getNode2() != null)
                    for (LTSNode n : nA) {
                        LTSTransition lt = new LTSTransition();
                        lt.setName(t.getName());
                        lt.setNode1(new LTSNode("(" + n.getName() + "," + t.getNode1().getName() + ")"));
                        lt.setNode2(new LTSNode("(" + n.getName() + "," + t.getNode2().getName() + ")"));
                        theNewLTS.addTransition(lt);
                    }
            }
        }
        
        // Lastly, set the start node
        theNewLTS.setStartNode(new LTSNode("(" + lts_a.getStartNode().getName() + "," + lts_b.getStartNode().getName() + ")"));
        
        
        return theNewLTS;
    }
    /** This converts the String syncSet to a HashSet and passes it to the normal compose function.
     * @returns The composition of lts_a and lts_b using the given Sync Set.
     */
    public static LTS composeA(LTS lts_a, LTS lts_b, String syncSet) {
        
        if (syncSet == "")
            return compose(lts_a,lts_b,null);
        HashSet<String> hs = new HashSet<String>();
        
        // First we must scan through and strip out any bracketed items:
        if (syncSet.contains("(") && syncSet.contains(")")) {
            // There are brackets. We must strip these items out first.
            String parts = "";
            boolean inBrackets = false;
            int startPosition = 0;
            for (int i = 0; i < syncSet.length(); i++) {
                if (syncSet.charAt(i) == '(')  {
                    inBrackets = true;
                    startPosition = i-1;
                    
                }
                if (inBrackets)                         parts += syncSet.charAt(i);
                if (syncSet.charAt(i) == ')')  {
                    inBrackets = false;
                    hs.add(parts);
                    parts = "";
                    String tmp = syncSet;
                    syncSet = tmp.substring(0,startPosition) + tmp.substring(i+1);
                }
            }
        }
        
        // Convert the string into a HashSet<String>:
        String[] result = syncSet.split(",");
        
        
        for (String s:result) {
            hs.add(s);
        }
        
        return compose(lts_a,lts_b,hs);
    }
    
    public static LTS Parallel_Reach(LTS lts_a, LTS lts_b, HashSet<String> syncSet) {
        
        // Cant have a null syncSet
        if (syncSet == null) syncSet = new HashSet<String>();
        
        
        // Prepare the parameters:
        parallel_reach_params params = new parallel_reach_params();
        params.syncSet = syncSet;
        params.nodeMapA = buildNodeMap(lts_a);
        params.nodeMapB = buildNodeMap(lts_b);
        params.ltsA = lts_a;
        params.ltsB = lts_b;
        params.transitionMapA = new Hashtable<String,LTSTransition>();
        params.transitionMapB = new Hashtable<String,LTSTransition>();
        params.newLTS = new LTS();
        params.nodesA = lts_a.getNodes(false);
        params.nodesB = lts_b.getNodes(false);
        params.transitionsA = lts_a.getTransitions();
        params.transitionsB = lts_b.getTransitions();
        params.visitedNodes = new HashSet<LTSNode>();
       
        for (LTSTransition t: params.transitionsA) {
            params.transitionMapA.put(t.getName(),t);
        }
        for (LTSTransition t: params.transitionsB) {
            params.transitionMapB.put(t.getName(),t);
        }
        
        
        // And now we do the composing stuff.
        doReachStuff(lts_a.getStartNode(), lts_b.getStartNode(), params);
        
        return params.newLTS;
    }
    
    private static void doReachStuff(LTSNode posA, LTSNode posB, parallel_reach_params params) {
        
        // We dont want to get stuck in a loop! make sure we havent been here before:
        LTSNode tmp = new LTSNode("(" + posA.getName() + "," + posB.getName() + ")");
        if (params.visitedNodes.contains(tmp)) return;
        else params.visitedNodes.add(tmp);
        
        // Transitions we can compose
        HashSet<LTSTransition> tA = params.nodeMapA.get(posA);
        HashSet<LTSTransition> tB = params.nodeMapB.get(posB);
        
        // Nodes we can compose with
        HashSet<LTSNode> pA = new HashSet<LTSNode>();
        HashSet<LTSNode> pB = new HashSet<LTSNode>();
        
        pB.add(posB);
        if (tB != null)
        for(LTSTransition t: tB) {
            if (t.getNode2() != null)
                pB.add(t.getNode2());
        }

        pA.add(posA);
        if (tA != null)
        for(LTSTransition t: tA) {
            if (t.getNode2() != null)
                pA.add(t.getNode2());
        }
        
        Hashtable<String,LTSTransition> htA = new Hashtable<String,LTSTransition>();
        Hashtable<String,LTSTransition> htB = new Hashtable<String,LTSTransition>();
        if (tB != null)
        for (LTSTransition t:tB) {
            htB.put(t.getName(),t);
        }
        if (tA != null)
        for (LTSTransition t:tA) {
            htA.put(t.getName(),t);
        }
        
        // Do the first lot of transitions and nodes
        if (tA != null) {
             // Now compute the transitions for the transitions from the first LTS and the nodes from the second LTS
            for (LTSTransition t: tA) {
                if (params.syncSet.contains(t.getName()) && htB.containsKey(t.getName())) {
                    LTSTransition b = htB.get(t.getName());

                    LTSTransition lt = new LTSTransition();
                    lt.setName(t.getName());
           
                    LTSNode nn = new LTSNode("(" + t.getNode1().getName() + "," + htB.get(t.getName()).getNode1().getName() + ")");
                    if (params.newLTS.getStartNode() == null) params.newLTS.setStartNode(nn); 
                    lt.setNode1(nn);
                    
                    lt.setNode2(new LTSNode("(" + t.getNode2().getName() + "," + htB.get(t.getName()).getNode2().getName() + ")"));
                    params.newLTS.addTransition(lt);
                    
                    doReachStuff(t.getNode2(), htB.get(t.getName()).getNode2(), params);
                    
                    continue;
                }
                else {
                    if (t.getNode2() != null)
                        for (LTSNode n:pB) {
                            LTSTransition lt = new LTSTransition();
                            lt.setName(t.getName());
                            LTSNode nn = new LTSNode("(" + t.getNode1().getName() + "," + n.getName() + ")");
                            if (params.newLTS.getStartNode() == null) params.newLTS.setStartNode(nn); 
                            lt.setNode1(nn);
                            lt.setNode2(new LTSNode("(" + t.getNode2().getName() + "," + n.getName() + ")"));
                            params.newLTS.addTransition(lt);

                           // Go go recursive stuff! if this were scheme we would need more brackets()
                           doReachStuff(t.getNode2(), n, params);
                        }
                }
            }
        }
        
        // Cant compose with nothing...
        if (tB == null) return;
       
        // And the second lot of transitions and nodes.      
        for (LTSTransition t: tB) {
             if (params.syncSet.contains(t.getName()) && htA.containsKey(t.getName())) {
                 // we have already synced.
                 continue;
             }
             else {
                if (t.getNode2() != null)
                    for(LTSNode n:pA) {
                        LTSTransition lt = new LTSTransition();
                        lt.setName(t.getName());
                        lt.setNode1(new LTSNode("(" + n.getName() + "," + t.getNode1().getName() + ")"));
                        lt.setNode2(new LTSNode("(" + n.getName() + "," + t.getNode2().getName() + ")"));
                        params.newLTS.addTransition(lt);

                        // Go go recursive stuff
                        doReachStuff(n, t.getNode2(), params);
                    }
             }
            }
    }
    
    /** Builds a HashMap using the first node as the key and a HashSet of transitions with
     * that node as the start node for the value.
     */
    private static HashMap<LTSNode, HashSet<LTSTransition>> buildNodeMap(LTS theLTS) {
        HashMap<LTSNode, HashSet<LTSTransition>> tmp = new HashMap<LTSNode, HashSet<LTSTransition>>();
        
        for (LTSTransition t:theLTS.getTransitions()) {
            
            if (!tmp.containsKey(t.getNode1())){
                tmp.put(t.getNode1(), new HashSet<LTSTransition>());
            }
            
            tmp.get(t.getNode1()).add(t);
        }
        
        return tmp;
    }
    
    /** Convenience overload function: does parallel reach using a string
     * as the sync set instead of a hash set.
     */
    public static LTS Parallel_ReachA(LTS lts_a, LTS lts_b, String syncSet) {
        
        if (syncSet == "")
            return compose(lts_a,lts_b,null);
        HashSet<String> hs = new HashSet<String>();
        
        // First we must scan through and strip out any bracketed items:
        if (syncSet.contains("(") && syncSet.contains(")")) {
            // There are brackets. We must strip these items out first.
            String parts = "";
            boolean inBrackets = false;
            int startPosition = 0;
            for (int i = 0; i < syncSet.length(); i++) {
                if (syncSet.charAt(i) == '(')  {
                    inBrackets = true;
                    startPosition = i-1;
                    
                }
                if (inBrackets)                         parts += syncSet.charAt(i);
                if (syncSet.charAt(i) == ')')  {
                    inBrackets = false;
                    hs.add(parts);
                    parts = "";
                    String tmp = syncSet;
                    syncSet = tmp.substring(0,startPosition) + tmp.substring(i+1);
                }
            }
        }
        
        // Convert the string into a HashSet<String>:
        String[] result = syncSet.split(",");
        
        
        for (String s:result) {
            hs.add(s);
        }
        
        return Parallel_Reach(lts_a,lts_b,hs);
    }
    
    /** Returns the name of the file this transition was loaded from or saved to.
     */
    public String getFileName() {
        return fileName;
    }
   
    /* This is a function for debugging purposes*/
    public static void printLTS(LTS theLTS) {
        System.out.println("======Dumping LTS======");
        System.out.println("StartNode: " + theLTS.getStartNode().toString());
        System.out.println("Transition Count: " + theLTS.getTransitions().size());
        System.out.println("Transitions: --");
        
        Collection<LTSTransition> ltst = theLTS.getTransitions();
        
        for (LTSTransition lt : ltst) {
            System.out.println("   " + lt.toString());
        }
        System.out.println("===LTS Dump Complete===");
        
    }
    
    
}
