/*
 * MainWindow.java
 *
 * Created on 23 July 2007, 21:07
 *
 */

package assignment;

/* Required packages */
import java.lang.*;
import java.io.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.JFileChooser;
import java.util.prefs.*;


/** Program entry point and main window class.
 *
 * @author David
 */
public class MainWindow implements ActionListener {
    
    protected JFrame window;        // The main window.
    protected LTSView theView;      // Displays an LTS
    protected LTS currentLTS;       // The currently open LTS.
    protected Preferences prefs;    // Application preferences.
    protected JFileChooser jfc;     // The file chooser we will use
    
    /**
     * Creates a new instance of MainWindow
     */
    public MainWindow(Preferences prefs) {
        this.prefs = prefs;
        
        window = new JFrame("COMP242 Assignment 0.2");
        window.setSize(800,600);
        window.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        jfc = new JFileChooser();
        Menu theMenu = new Menu(this, prefs);

        window.setLayout(new BorderLayout());
        
        window.setJMenuBar(theMenu);
        
        
        // Create a new LTS View panel.
        theView = new ForceLTSView();
        window.add(theView, BorderLayout.CENTER);
        
        JToolBar tb = new JToolBar("Exciting buttons of doom");
        addTBButtons(tb);
        window.add(tb,BorderLayout.PAGE_START);
        
        
        window.setVisible(true);
        
        System.out.println("Autocreating LTS");
        fileNew();
    }
    
    /** Adds the buttons to the toolbar
     */
    public void addTBButtons(JToolBar tb) {
        JButton button = null;
        
        button = new JButton();
        button.setText("Open");
        button.addActionListener(this);
        tb.add(button);
        
        button = new JButton();
        button.setText("Compose");
        button.addActionListener(this);
        tb.add(button);
        
        button = new JButton();
        button.setText("Reach");
        button.addActionListener(this);
        tb.add(button);
        
        button = new JButton();
        button.setText("Redraw");
        button.addActionListener(this);
        tb.add(button);
        
        button = new JButton();
        button.setText("Iterate");
        button.addActionListener(this);
        tb.add(button);
        
        button = new JButton();
        button.setText("Reset Layout");
        button.addActionListener(this);
        tb.add(button);
    }
    
    /** Handles the toolbar events
     */
    public void actionPerformed(ActionEvent event){
        String s = event.getActionCommand();
        
        if (s == "Open")
            handleMenuEvent("Open...");
        if (s == "Compose")
            handleMenuEvent("Compose with...");
        if (s == "Reach")
            handleMenuEvent("Parallel Reach...");
        if (s == "Parallel Reach...")
            handleMenuEvent("Parallel Reach...");
        if (s == "Redraw")
            handleMenuEvent("Refresh");
        if (s == "Iterate")
            handleMenuEvent("Sim Iterate");
        if (s == "Reset Layout")
            handleMenuEvent(s);
     }
    
    /** Processes a menu click event.
     */
    public void handleMenuEvent(String menuItemPressed) {
        if (menuItemPressed == "New")
            fileNew();
        else if (menuItemPressed == "Open...")
            fileOpen();
        else if (menuItemPressed == "Save")
            fileSave();
        else if (menuItemPressed == "Save As...")
            fileSaveAs();
        else if (menuItemPressed == "Compose with...")
            editCompose();
        else if (menuItemPressed == "Parallel Reach...")
            editParallelReach();
        else if (menuItemPressed == "Exit")
            exit();
        else if (menuItemPressed == "Refresh")
            updateDisplay();
        else if (menuItemPressed == "Sim Iterate")
            iterate();
        else if (menuItemPressed == "Reset Layout")
            resetLayout();
        else             
            System.out.println("Error: Unknown menu event name \"" + menuItemPressed + "\"");
    }
    
    /** Switches to the specified tool
     */
    public void switchTool(String ToolName) {
        /* Tools for the Viewer/Editor are:
         * 0 - No tool
         * 1 - Select
         * 2 - Set Start Node
         * 3 - Add Node
         * 4 - Add Transition
         * 5 - Remove Node
         * 6 - Remove Transition
         * 7 - Move Node
         */
        
        if (ToolName == "Select")
            theView.setTool(1);
        else if (ToolName == "Set Start Node")
            theView.setTool(2);
        else if (ToolName == "Add Node")
            theView.setTool(3);
        else if (ToolName == "Add Transition")
            theView.setTool(4);
        else if (ToolName == "Remove Node")
            theView.setTool(5);
        else if (ToolName == "Remove Transition")
            theView.setTool(6);
        else if (ToolName == "Move Node")
            theView.setTool(7);
        else
            theView.setTool(0);
    }
    
    /** This is for future use by a most recently used file menus
     */
    public void handleOpenEvent(String fileName) {
        try {
            currentLTS = LTS.load(fileName);

            updateTitle(fileName);

            theView.setLTS(currentLTS);
            theView.setup();
                        }
        catch (IOException e) {
            System.out.println("File load failed: IO Exception");
        }
    }
    
    /** Creates a new LTS.
     */
    private void fileNew() {
        System.out.println("Creating new LTS...");
        currentLTS = new LTS();
        updateTitle("Untitled");
        theView.setLTS(currentLTS);
        theView.setup();
        //updateDisplay();
    }
    
    /** Composes the current LTS with another. First asks the user to select the
     * file containing the LTS to compose with the current one. Then asks the
     * user to enter the Sync Set.
     */
    private void editCompose() {
        LTS ltsA = currentLTS;
        if (!fileOpen()) {
            // Open operation canceled: die.
            System.out.println("Compose Canceled by user");
            return;
        }
        currentLTS.setFileName(ltsA.getFileName());
        updateTitle(currentLTS.getFileName());
        LTS ltsB = currentLTS;
        
        // Ok, now we need to get the Sync set.
        String syncSet = JOptionPane.showInputDialog("Enter Sync Set:");
        
        currentLTS = LTS.composeA(ltsA, ltsB, syncSet);
        currentLTS.setFileName(ltsA.getFileName());
        updateTitle(currentLTS.getFileName());
        theView.setLTS(currentLTS);
        theView.setup();
    }
    
    /** Parallel Reach composes the current LTS with another. First asks the 
     * user to select the file containing the LTS to compose with the current 
     * one. Then asks the user to enter the Sync Set.
     */
    private void editParallelReach() {
        LTS ltsA = currentLTS;
        if (!fileOpen()) {
            // Open operation canceled: die.
            System.out.println("Parallel Reach Canceled by user");
            return;
        }
        currentLTS.setFileName(ltsA.getFileName());
        updateTitle(currentLTS.getFileName());
        LTS ltsB = currentLTS;
        
        // Ok, now we need to get the Sync set.
        String syncSet = JOptionPane.showInputDialog("Enter Sync Set:");
        
        currentLTS = LTS.Parallel_ReachA(ltsA, ltsB, syncSet);
        currentLTS.setFileName(ltsA.getFileName());
        currentLTS.setFileName(ltsA.getFileName());
        updateTitle(currentLTS.getFileName());
        theView.setLTS(currentLTS);
        theView.setup();
    }
    
    /** Opens an LTS from a text file.
     */
    private boolean fileOpen() {
        int returnVal = jfc.showOpenDialog(jfc);
       
        if (returnVal == JFileChooser.APPROVE_OPTION) {
            /* Load the file */
            File file = jfc.getSelectedFile();
            
            /*prefs.put("recentFiles/" + file.getName(), file.getAbsolutePath());
            
            try {
                prefs.flush();
            } catch (Exception e) {};*/
            
            handleOpenEvent(file.getAbsolutePath());
            
            return true;
            
            /* Update the display */
            //updateDisplay();
        }
        else {
            System.out.println("File open operation canceled by user.");
            return false;
        }
    }
    
    /** Saves the current LTS.
     */
    private void fileSave() {
        try {
            theView.flushChanges();
            currentLTS.save();
        }
        catch (IOException e) {
            System.out.println("File Save failed: " + e.toString());
        }
    }
    
    /** Saves the current LTS with a different file name.
     */
    private void fileSaveAs() {
        int returnVal = jfc.showSaveDialog(jfc);
       
        if (returnVal == JFileChooser.APPROVE_OPTION) {
            File file = jfc.getSelectedFile();
            try {
                theView.flushChanges();
                currentLTS.save(file.getAbsolutePath());
                updateTitle(file.getAbsolutePath());
            }
            catch (IOException e) {
                System.out.println("File Save failed: " + e.toString());
            }
        } else {
            System.out.println("File save operation canceled by user.");
        }
    }
     
    /** Exits the program.
     */
    private void exit() {
        System.out.println("Application terminated by user.");
        System.exit(0);     /* Terminate Normally */
    }
    
    /** Resets the layout so that it may be rerun if the current layout
     *  isnt ideal.
     */
    private void resetLayout() {
        //window.remove(theView);
        //theView = new ForceLTSView();
        //theView.setLTS(currentLTS);
        //window.add(theView, BorderLayout.CENTER);
        theView.setup();
        theView.repaint();
    }
    
    /** Updates the display (recalculates/redraws the LTS)
     */
    private void updateDisplay() {
        System.out.println("Updating display...");
        theView.renderLTS();
        System.out.println("Display updated.");
    }
    
    /** Updates the string displayed in the title bar
     */
    private void updateTitle(String addition) {
        window.setTitle("COMP242 Assignment 0.2 - [" + addition + "]");
    }
    
    /** Does one iteration of the layout algorithm
     */
    private void iterate() {
        Double val = theView.iterate();
        updateTitle(val.toString());
    }
    
    /** Program entry point.
     * @param args the command line arguments
     */
    public static void main(String[] args) {
        System.out.println("COMP242 Assignment version 0.2\n     (C) Copyright David Goodwin, 2007");
        System.out.println("\nProgram Starting...");
        
        // Load preferences:
        //Preferences prefs = Preferences.userRoot().node("/com/drg/242a");
        
        MainWindow theApp = new MainWindow(null);
    }
    
}
